package com.flyfish.framework.logging.advice;

import com.flyfish.framework.logging.service.LogContext;
import com.flyfish.framework.logging.service.LogManager;
import lombok.RequiredArgsConstructor;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.reactivestreams.Publisher;
import org.reactivestreams.Subscriber;
import org.reactivestreams.Subscription;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

/**
 * 日志切面
 *
 * @author wangyu
 */
@Aspect
@RequiredArgsConstructor
public class LogAdvice {

    private final LogManager logManager;

    /**
     * 切入注解
     */
    @Pointcut("execution(public * com.flyfish..*.*Controller.*(..))")
    public void pointCut() {

    }


    /**
     * 环绕切面，获取参数
     *
     * @param joinPoint 切入点
     * @return 结果
     */
    @Around("pointCut()")
    public Object process(ProceedingJoinPoint joinPoint) throws Throwable {
        // 构建上下文
        LogContext context = LogContext.of(joinPoint);
        // 结果信息
        Object result = null;
        try {
            return result = joinPoint.proceed(joinPoint.getArgs());
        } catch (Throwable throwable) {
            context.setError(throwable);
            throw throwable;
        } finally {
            if (context.isValid()) {
                handleResult(result, context);
                logManager.tryLog(context);
            }
        }
    }

    /**
     * 处理结果，支持webflux
     *
     * @param result  结果
     * @param context 上下文
     */
    private void handleResult(Object result, LogContext context) {
        if (null != result) {
            // 判断结果是否是publisher
            if (result instanceof Publisher) {
                if (result instanceof Mono) {
                    ((Mono<?>) result).subscribe(context::setResult, context::setError, context::end);
                } else if (result instanceof Flux) {
                    ((Flux<?>) result).collectList().subscribe(context::setResult, context::setError, context::end);
                } else {
                    ((Publisher<?>) result).subscribe(new LogSubscriber(context));
                }
            } else {
                context.setResult(result);
                context.end();
            }
        } else {
            // 为空，直接标记结束
            context.end();
        }
    }

    @RequiredArgsConstructor
    @SuppressWarnings("all")
    private static class LogSubscriber implements Subscriber<Object> {

        private final LogContext context;

        @Override
        public void onSubscribe(Subscription s) {

        }

        @Override
        public void onNext(Object o) {
            context.setResult(o);
        }

        @Override
        public void onError(Throwable t) {
            context.setError(t);
        }

        @Override
        public void onComplete() {
            context.end();
        }
    }
}
